// TEXAS INSTRUMENTS TEXT FILE LICENSE
//
//
// Copyright (c) 2020 Texas Instruments Incorporated
//
// All rights reserved not granted herein.
//
// Limited License.
//
// Texas Instruments Incorporated grants a world-wide, royalty-free, non-exclusive license under copyrights
// and patents it now or hereafter owns or controls to make, have made, use, import, offer to sell and sell
// ("Utilize") this software subject to the terms herein.  With respect to the foregoing patent license,
// such license is granted solely to the extent that any such patent is necessary to Utilize the software
// alone.  The patent license shall not apply to any combinations which include this software, other than
// combinations with devices manufactured by or for TI ("TI Devices").  No hardware patent is licensed
// hereunder.
//
// Redistributions must preserve existing copyright notices and reproduce this license (including the above
// copyright notice and the disclaimer and (if applicable) source code license limitations below) in the
// documentation and/or other materials provided with the distribution
//
// Redistribution and use in binary form, without modification, are permitted provided that the following
// conditions are met:
//
// *  No reverse engineering, decompilation, or disassembly of this software is permitted with respect to
// any software provided in binary form.
//
// * Nothing shall obligate TI to provide you with source code for the software licensed and provided to
// you in object code.
//
// If software source code is provided to you, modification and redistribution of the source code are
// permitted.
//
// Neither the name of Texas Instruments Incorporated nor the names of its suppliers may be used to
// endorse or promote products derived from this software without specific prior written permission.
//
// DISCLAIMER.
//
// THIS SOFTWARE IS PROVIDED BY TI AND TI'S LICENSORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL TI AND TI'S LICENSORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
// OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
// AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.


#include <Wire.h>
#include <aJSON.h>

// Constants
const uint16_t SAMPLE_PERIOD = 100;     // Sample/PID period in ms
const uint16_t GUI_DATA_PERIOD = 200;   // Transmission to GUI period in ms
const float TEMP_GOAL = 25;             // Initial goal temperature
const uint8_t VOUT_INIT = 60;           // Initial output voltage, 1.8 V + 60 * 0.025 V = 3.3 V
const uint8_t VOUT_MAX = 136;           // Maximum output voltage, 1.8 V + 136 * 0.025 V = 5.2 V
const uint8_t VOUT_MIN = 0;             // Minimum output voltage, 1.8 V + 0 * 0.025 V = 1.8 V
const float PID_KP = 5;                 // Initial PID parameters
const float PID_KI = 0.0005;
const float PID_KD = 0;
const float PID_KAWU = 10;
const float VOUT_OFFSET = 1.8;          // Offset value for VOUT
const float VOUT_GAIN = 40.0;           // Gain value for VOUT

typedef enum {OPEN_LOOP, CLOSED_LOOP};        // Control modes
typedef enum {SENS_TYPE_I2C, SENS_TYPE_ANL};  // Sensor types
typedef enum {
    DATA_TEMP_SENS,
    DATA_TEMP_GOAL,
    DATA_SENS_TYPE,
    DATA_KP,
    DATA_KI,
    DATA_KD,
    DATA_KAWU,
    DATA_VOUT,
    DATA_CTRL
};  // LP to GUI messages

const uint8_t TMP117_ADDRESS = 0x48;    // TMP117 info
typedef enum  {
    TMP117_TEMP_REG = 0,
    TMP117_CONFIG_REG
};

const uint8_t TPS63810_ADDRESS = 0x75;  // TPS63810 info
typedef enum {
    TPS63810_CONTROL_REG = 1,
    TPS63810_STATUS_REG,
    TPS63810_DEVID_REG,
    TPS63810_VOUT1_REG,
    TPS63810_VOUT2_REG
};

// Variables
uint8_t control = OPEN_LOOP;        // Control mode
uint8_t sensType = SENS_TYPE_I2C;   // Temperaure sensor type
float tempSens = 0;                 // Temperature read by sensor        
float tempGoal = TEMP_GOAL;         // Goal temperature
float tempErr = 0;                  // Temperature error
float pidIntegral = 0;              // PID controller integrator
float pidKp = PID_KP;               // PID controller proportional gain
float pidKi = PID_KI;               // PID controller integral gain
float pidKd = PID_KD;               // PID controller derivative gain
float pidKawu = PID_KAWU;           // PID anti-windup gain
float pidAntiWU = 0;                // 
float voutPID = VOUT_INIT;          // PID controller output
float vout = VOUT_INIT;             // PID controller limited output - value sent to TPS63810

uint32_t tmrMain = 0;               // Timers for periodic tasks
uint32_t tmrData = 0;

aJsonStream serial_stream(&Serial); // JSON serial stream

// Function prototypes
void send_to_GUI(int param, float temp);
void processMessage(aJsonObject *msg);

void setup() {
  // Configure communication interfaces
  Serial.begin(57600);                      // Start serial port
  Wire.begin();                             // Start I2C
  
  // Configure ADC
  analogReference(INTERNAL2V5);             // !! Launchpad specific !! Set ADC reference to EXTERNAL,
//  pinMode(37, INPUT);                       // !! Launchpad specific !! Set Veref+ pin type to INPUT, 

  // Configure TPS63811
  Wire.beginTransmission(TPS63810_ADDRESS);
  Wire.write(TPS63810_CONTROL_REG);         // CONTROL_REGISTER:
  Wire.write(0b00101000);                   // RANGE = 0, ENABLE = 1, FPWM = 1, RPWM = 0, SLEW = 00 
  Wire.endTransmission();
  Wire.beginTransmission(TPS63810_ADDRESS);
  Wire.write(TPS63810_VOUT1_REG);           // VOUT1
  Wire.write(VOUT_INIT);                    // = VOUT_INIT 
  Wire.endTransmission();
  
  // Configure I2C Sensor
  if(sensType == SENS_TYPE_I2C) {
    Wire.beginTransmission(TMP117_ADDRESS);
    Wire.write(TMP117_CONFIG_REG);          // CONFIGURATION_REGISTER:
    Wire.write(0x00);                       // 0x00, continuous conversion, no averaging
    Wire.write(0x00);                       // 0x00 
    Wire.endTransmission();
    Wire.beginTransmission(TMP117_ADDRESS);
    Wire.write(TMP117_TEMP_REG);            // Set pointer to TEMP_RESULT_REGISTER
    Wire.endTransmission();
  }

  delay(3140);                              // Wait for 3.14 s to stabilize

  // Initialize GUI values
  send_to_GUI(DATA_TEMP_GOAL, tempGoal); 
  send_to_GUI(DATA_CTRL, control); 
  send_to_GUI(DATA_VOUT, VOUT_OFFSET + 1/VOUT_GAIN * VOUT_INIT); 
  send_to_GUI(DATA_KP, pidKp); 
  send_to_GUI(DATA_KI, pidKi); 
  send_to_GUI(DATA_KD, pidKd); 
  send_to_GUI(DATA_KAWU, pidKawu); 
}

void loop() {
  uint16_t tmp = 0;

  // Main control loop, period defined by SAMPLE_PERIOD
  if (millis() - tmrMain >= SAMPLE_PERIOD){
    tmrMain = millis();

    // Read temperature sensor
    switch (sensType) {
      // I2C Sensor readout - TMP117
      case SENS_TYPE_I2C:
        Wire.requestFrom(TMP117_ADDRESS,2);
        while(Wire.available())  
          tmp = (tmp << 8) | Wire.read();
        tempSens = tmp*0.0078125;
        break;
      // Analog sensor readout
      case SENS_TYPE_ANL:                               
        tmp = analogRead(6);                            // Read analog input at BoosterPack Pin 6
        //tempSens = 29 + 0.16667 * (tmp - 2091.0);     // TMP6131 conversion, Rhs = 10k, Vbias = 1.225 V (LM4041), ADCref = 1.2 V, resolution = 1/6 C
        tempSens = 25 - 0.023923 * (tmp - 2091.0);      // 10k NTC conversion, Rhs=10k, Vbias=1.225V (LM4041), ADCref=1.2V
        break;
      default:
        break;      
    }
    
    // PID controller
    if (control == CLOSED_LOOP){
      tempErr = tempSens - tempGoal;                                    // Calculate temperature error
      pidIntegral = pidIntegral + tempErr * SAMPLE_PERIOD + pidAntiWU;  // Calculate integral
      voutPID = pidKp * tempErr + pidKi * pidIntegral;                  // Calculate PID output
      if (voutPID > VOUT_MAX)                                           // Limit PID controller output
        vout = VOUT_MAX;
      else if (voutPID < VOUT_MIN)
        vout = VOUT_MIN;
      else 
        vout = voutPID;
      pidAntiWU = -pidKawu * (voutPID - vout);                          // Anti windup
    }
  
    // Adjust output voltage
    if (vout > 127) {                             // If VOUT > 4.975 V switch to HIGH range
        Wire.beginTransmission(TPS63810_ADDRESS);
        Wire.write(TPS63810_CONTROL_REG);         // CONTROL register:
        Wire.write(0b01101000);                   // RANGE = 1, ENABLE = 1, FPWM = 1 
        Wire.endTransmission();
        Wire.beginTransmission(TPS63810_ADDRESS); // Update VOUT
        Wire.write(TPS63810_VOUT1_REG);
        Wire.write((uint8_t) vout - 9);
        Wire.endTransmission();
    } else {                                      // If VOUT <= 4.975 V switch to LOW range
        Wire.beginTransmission(TPS63810_ADDRESS);
        Wire.write(TPS63810_CONTROL_REG);         // CONTROL register:
        Wire.write(0b00101000);                   // RANGE = 0, ENABLE = 1, FPWM = 1 
        Wire.endTransmission();
        Wire.beginTransmission(TPS63810_ADDRESS); // Update VOUT
        Wire.write(TPS63810_VOUT1_REG);
        Wire.write((uint8_t) vout);
        Wire.endTransmission();
    }
  }

  // Send data to GUI periodically, period defined by GUI_DATA_PERIOD
  if (millis() - tmrData >= GUI_DATA_PERIOD){
    tmrData = millis();
    send_to_GUI(DATA_TEMP_SENS, tempSens); 
    if (control == CLOSED_LOOP) 
      send_to_GUI(DATA_VOUT, VOUT_OFFSET + 1/VOUT_GAIN * (uint8_t) vout); 
  }
     
  // Receive message from GUI
  if (serial_stream.available()) {
    aJsonObject *msg = aJson.parse(&serial_stream);
    processMessage(msg);
    aJson.deleteItem(msg);
  }
}

// Send data to GUI
// Parameters:
//   param: parameter type to be sent
//   val:   value of the parameter
void send_to_GUI(int param, float val) {
  aJsonObject* root = aJson.createObject();
  char* string = NULL;
  if (root == NULL) return;
  switch (param) {
    case DATA_TEMP_SENS:
      aJson.addItemToObject(root, "TEMP_SENS", aJson.createItem(val));
      break;
    case DATA_TEMP_GOAL:
      aJson.addItemToObject(root, "TEMP_GOAL", aJson.createItem(val));
      break;
    case DATA_SENS_TYPE:
      aJson.addItemToObject(root, "SENS_TYPE", aJson.createItem((int)val));
      break;
    case DATA_KP:
      aJson.addItemToObject(root, "PID_KP", aJson.createItem(val));
      break;
    case DATA_KI:
      aJson.addItemToObject(root, "PID_KI", aJson.createItem(val));
      break;
    case DATA_KD:
      aJson.addItemToObject(root, "PID_KD", aJson.createItem(val));
      break;
    case DATA_KAWU:
      aJson.addItemToObject(root, "PID_KAWU", aJson.createItem(val));
      break;
    case DATA_VOUT:
      aJson.addItemToObject(root, "VOUT", aJson.createItem(val));
      break;
    case DATA_CTRL:
      aJson.addItemToObject(root, "CTRL", aJson.createItem(val));
      break;
    default:
      break;
  }
  string = aJson.print(root);
  if (string != NULL) {
    Serial.println(string);
    free(string);
  } 
  aJson.deleteItem(root);
}

// Decode message from GUI
// Parameter:
//   *msg : pointer to received JSON object message
void processMessage(aJsonObject *msg) {
  aJsonObject *tmp = NULL;
  if (tmp = aJson.getObjectItem(msg, "CTRL"))
    if (tmp->type == aJson_Int)
      control = tmp->valueint;
  else
    return;    
  if (tmp = aJson.getObjectItem(msg, "TEMP_GOAL"))
    if (tmp->type == aJson_Int)
      tempGoal = tmp->valueint;
    else if (tmp->type == aJson_Float)
      tempGoal = tmp->valuefloat; 
  else
    return;    
  if (tmp = aJson.getObjectItem(msg, "PID_KP"))
    if (tmp->type == aJson_Int)
      pidKp = tmp->valueint;
    else if (tmp->type == aJson_Float)
      pidKp = tmp->valuefloat;
  else
    return;    
  if (tmp = aJson.getObjectItem(msg, "PID_KI"))
    if (tmp->type == aJson_Int)
      pidKi = tmp->valueint;
    else if (tmp->type == aJson_Float)
      pidKi = tmp->valuefloat;
  else
    return;    
  if (tmp = aJson.getObjectItem(msg, "PID_KD"))
    if (tmp->type == aJson_Int)
      pidKd = tmp->valueint;
    else if (tmp->type == aJson_Float)
      pidKd = tmp->valuefloat;
  else
    return;    
  if (tmp = aJson.getObjectItem(msg, "PID_KAWU"))
    if (tmp->type == aJson_Int)
      pidKawu = tmp->valueint;
    else if (tmp->type == aJson_Float)
      pidKawu = tmp->valuefloat;
  else
    return;    
  if (tmp = aJson.getObjectItem(msg, "VOUT"))
    if (tmp->type == aJson_Float) {
      vout = (tmp->valuefloat - VOUT_OFFSET) * VOUT_GAIN;
    }
  else
    return;    
  if (tmp = aJson.getObjectItem(msg, "INIT_TPS")) {
    control = CLOSED_LOOP;                        // Initialize PID
    send_to_GUI(DATA_TEMP_GOAL, tempGoal);        // Initialize GUI values
    send_to_GUI(DATA_KP, pidKp); 
    send_to_GUI(DATA_KI, pidKi); 
    send_to_GUI(DATA_KD, pidKd); 
    send_to_GUI(DATA_KAWU, pidKawu); 
    send_to_GUI(DATA_CTRL, control); 
    send_to_GUI(DATA_VOUT, VOUT_OFFSET + 1/VOUT_GAIN * vout); 
    
    Wire.beginTransmission(TPS63810_ADDRESS);     // Initalize TPS63810
    Wire.write(TPS63810_CONTROL_REG);             // CONTROL_REGISTER:
    Wire.write(0b00101000);                       // RANGE=0, ENABLE=1, FPWM=1, RPWM=0, SLEW=00 
    Wire.endTransmission();
  } else
    return;    
}
